/**
 * 
 */
package com.wieseke.cptk.common.viewer;

import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;

import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.util.Scaling;

/**
 * @author Nicolas Wieseke
 * 
 */
public abstract class NodeControl extends Canvas {

	/**
	 * the width of the node
	 */
	public int width;

	/**
	 * the height of the node
	 */
	public int height;
	/**
	 * the cophylogeny viewer
	 */
	protected CophylogenyViewer viewer;

	/**
	 * the corresponding node
	 */
	protected IPositionNode node;

	/**
	 * the parent nodes control
	 */
	protected NodeControl parentNodeControl;

	/**
	 * the child node controls
	 */
	protected List<NodeControl> childNodeControls;

	/**
	 * flag indicating if the node is folded
	 */
	protected boolean foldFlag;
	/**
	 * flag indicating if the mouse is over the node
	 */
	protected boolean mouseOverFlag;

	/**
	 * the currend scale factor
	 */
	protected double scaleFactor;

	/**
	 * default constructor
	 * 
	 * @param parent
	 * @param style
	 * @param node
	 */
	public NodeControl(CophylogenyViewer viewer, IPositionNode node,
			NodeControl parent) {
		super(viewer.container, SWT.TRANSPARENT);

		this.viewer = viewer;
		this.node = node;
		this.parentNodeControl = parent;
		this.childNodeControls = new ArrayList<NodeControl>();

		this.foldFlag = true;
		this.mouseOverFlag = false;
		this.scaleFactor = 1d;

		this.setToolTipText(node.getToolTipText());

		// add the paint listener
		addPaintListener(new PaintListener() {

			public void paintControl(PaintEvent e) {
				NodeControl.this.paintControl(e.gc);
			}
		});
		// add the mouse track listener
		addMouseTrackListener(new MouseTrackListener() {

			public void mouseEnter(MouseEvent e) {
				NodeControl.this.mouseOverFlag = true;
				// TODO find better way to update tooltips
				NodeControl.this.setToolTipText(NodeControl.this
						.getToolTipText());
				redraw();
				NodeControl.this.viewer.redraw();
			}

			public void mouseExit(MouseEvent e) {
				mouseOverFlag = false;
				redraw();
				NodeControl.this.viewer.redraw();
			}

			public void mouseHover(MouseEvent e) {
			}
		});
		// add the mouse listener
		addMouseListener(new MouseListener() {

			public void mouseDoubleClick(MouseEvent e) {
			}

			public void mouseDown(MouseEvent e) {
			}

			public void mouseUp(MouseEvent e) {
				if (e.button == 3) {
					cancelOperations();
				}
			}
		});
		// add the key listener
		addKeyListener(new KeyListener() {

			public void keyPressed(KeyEvent e) {
				if (e.character == '+') {
					NodeControl.this.viewer.zoomIn();
				} else if (e.character == '-') {
					NodeControl.this.viewer.zoomOut();
				} else if (e.keyCode == SWT.ARROW_UP) {
					Point oldOrigin = NodeControl.this.viewer.getOrigin();
					NodeControl.this.viewer.setOrigin(oldOrigin.x, Math.max(0,
							oldOrigin.y - 10));
				} else if (e.keyCode == SWT.ARROW_DOWN) {
					Point oldOrigin = NodeControl.this.viewer.getOrigin();
					NodeControl.this.viewer.setOrigin(oldOrigin.x, Math.min(
							NodeControl.this.viewer.container.getSize().y,
							oldOrigin.y + 10));
				} else if (e.keyCode == SWT.ARROW_LEFT) {
					Point oldOrigin = NodeControl.this.viewer.getOrigin();
					NodeControl.this.viewer.setOrigin(Math.max(0,
							oldOrigin.x - 10), oldOrigin.y);
				} else if (e.keyCode == SWT.ARROW_RIGHT) {
					Point oldOrigin = NodeControl.this.viewer.getOrigin();
					NodeControl.this.viewer.setOrigin(Math.min(
							NodeControl.this.viewer.container.getSize().y,
							oldOrigin.x + 10), oldOrigin.y);
				}
			}

			public void keyReleased(KeyEvent e) {
				if (e.keyCode == SWT.ESC) {
					cancelOperations();
				}
			}
		});
		addFocusListener(new FocusListener() {

			public void focusLost(FocusEvent e) {
				redraw();
				NodeControl.this.viewer.redraw();
			}

			public void focusGained(FocusEvent e) {
				redraw();
				NodeControl.this.viewer.redraw();
			}
		});
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.swt.widgets.Control#getToolTipText()
	 */

	public String getToolTipText() {
		return this.node.getToolTipText();
	}

	/**
	 * paint the control
	 * 
	 * @param gc
	 */
	protected abstract void paintControl(GC gc);

	/**
	 * paint the control
	 * 
	 * @param gc
	 */
	protected abstract void paintControl(Graphics2D g2d);

	/**
	 * scale an integer to the current scale factor
	 * 
	 * @param value
	 * @return the scaled integer
	 */
	public int scale(int value) {
		return Scaling.convertToScaled(value, this.scaleFactor);
	}

	/**
	 * normalize an integer according to the current scale factor
	 * 
	 * @param value
	 * @return the normalized integer
	 */
	public int normalize(int value) {
		return Scaling.convertFromScaled(value, this.scaleFactor);
	}

	/**
	 * set the position for the underlying node
	 * 
	 * @param x
	 * @param y
	 * @param refresh
	 *            flag indicating if the visual component should be set to the
	 *            new coordinates
	 */
	public void setPosition(int x, int y, boolean refresh) {
		this.node.setPosX(x);
		this.node.setPosY(y);
		if (refresh) {
			int posX = scale(x);
			int posY = scale(y);
			if (posX != getLocation().x || posY != getLocation().y)
				setLocation(posX, posY);
		} else
			this.redraw();
	}

	/**
	 * apply the position and size and force a redraw
	 */
	public void applyPosition() {
		int posX = scale(this.node.getPosX());
		int posY = scale(this.node.getPosY());
		int width = scale(getNormalizedSize().x);
		int height = scale(getNormalizedSize().y);
		if (posX != getLocation().x || posY != getLocation().y
				|| width != getSize().x || height != getSize().y)
			setBounds(posX, posY, width, height);
	}

	/**
	 * toggle the expand flag and hide or show the nodes in the subtree
	 * recursively
	 */
	public void toggleExpandFlag() {
		if (!node.isLeaf()) {
			this.foldFlag = !this.foldFlag;
		}
		if (!this.foldFlag)
			for (NodeControl childNode : this.childNodeControls) {
				childNode.hideSubtree();
			}
		else
			for (NodeControl childNode : this.childNodeControls) {
				childNode.showSubtree();
			}
	}

	/**
	 * hide the subtree inclusive the current node
	 */
	public void hideSubtree() {
		for (NodeControl childNode : this.childNodeControls) {
			childNode.hideSubtree();
		}
		this.setVisible(false);
	}

	/**
	 * show the subtree inclusive the current node
	 */
	public void showSubtree() {
		if (this.foldFlag)
			for (NodeControl childNode : this.childNodeControls) {
				childNode.showSubtree();
			}
		this.setVisible(true);
		this.redraw();
	}

	/**
	 * flip the node horizontal in the given rectangle
	 * 
	 * @param rect
	 */
	public void flipHorizontal(Rectangle rect) {
		setPosition(rect.x + (rect.width - this.node.getPosX())
				- getNormalizedSize().x, this.node.getPosY(), true);
		applyPosition();
	}

	/**
	 * flip the node vertical in the given rectangle
	 * 
	 * @param rect
	 */
	public void flipVertical(Rectangle rect) {
		setPosition(this.node.getPosX(), rect.y
				+ (rect.height - this.node.getPosY()) - getNormalizedSize().y,
				true);
		applyPosition();
	}

	/**
	 * switch x and y position of the node in the given rectangle
	 * 
	 * @param rect
	 */
	public void rotate(Rectangle rect) {
		setPosition(this.node.getPosY() + (rect.x - rect.y), this.node
				.getPosX()
				+ (rect.y - rect.x), true);
		applyPosition();
	}

	/**
	 * cancel all not finished operations when ESC or the right mouse button is
	 * pressed
	 */
	public abstract void cancelOperations();

	/**
	 * @return the node
	 */
	public IPositionNode getNode() {
		return node;
	}

	/**
	 * @param node
	 *            the node to set
	 */
	public void setNode(IPositionNode node) {
		this.node = node;
	}

	/**
	 * @return the the normalized x position of the node
	 */
	public int getPosX() {
		return this.node.getPosX();
	}

	/**
	 * @return the the normalized y position of the node
	 */
	public int getPosY() {
		return this.node.getPosY();
	}

	/**
	 * @return the scaleFactor
	 */
	public double getScaleFactor() {
		return scaleFactor;
	}

	/**
	 * @param scaleFactor
	 *            the scaleFactor to set
	 */
	public void setScaleFactor(double scaleFactor) {
		this.scaleFactor = scaleFactor;
	}

	/**
	 * @return the mouseOverFlag
	 */
	public boolean isMouseOverFlag() {
		return mouseOverFlag;
	}

	/**
	 * @return the parentNodeControl
	 */
	public NodeControl getParentNodeControl() {
		return parentNodeControl;
	}

	/**
	 * @param parentNodeControl
	 *            the parentNodeControl to set
	 */
	public void setParentNodeControl(NodeControl parentNodeControl) {
		this.parentNodeControl = parentNodeControl;
	}

	/**
	 * @return the childNodeControls
	 */
	public List<NodeControl> getChildNodeControls() {
		return childNodeControls;
	}

	/**
	 * @param childNodeControls
	 *            the childNodeControls to set
	 */
	public void setChildNodeControls(List<NodeControl> childNodeControls) {
		this.childNodeControls = childNodeControls;
	}

	/**
	 * @return the normalized size of the control
	 */
	protected abstract Point getNormalizedSize();

}
