/**
 * 
 */
package com.wieseke.cptk.reconstruction.dao;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;

import com.wieseke.cptk.common.api.ICophylogeny;
import com.wieseke.cptk.common.constants.EventConstants;
import com.wieseke.cptk.common.constants.OptionsConstants;
import com.wieseke.cptk.common.dao.Cophylogeny;
import com.wieseke.cptk.common.dao.ReconstructionCosts;
import com.wieseke.cptk.common.log.CophylogenyLogger;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;
import com.wieseke.cptk.output.editor.VisualOutputEditor;
import com.wieseke.cptk.output.editor.VisualOutputEditorInput;
import com.wieseke.cptk.reconstruction.solver.BaseSolver;
import com.wieseke.cptk.reconstruction.util.CostCalculator;
import com.wieseke.cptk.reconstruction.util.QualityCalculator;
import com.wieseke.cptk.reconstruction.util.RestrictedQualityCalculator;

import de.jtem.numericalMethods.calculus.function.RealFunctionOfSeveralVariables;
import de.jtem.numericalMethods.calculus.minimizing.Info;
import de.jtem.numericalMethods.calculus.minimizing.NelderMead;

/**
 * @author Nicolas Wieseke
 * 
 */
public class ReconstructionCophylogeny extends Cophylogeny implements
		ICophylogeny, RealFunctionOfSeveralVariables {

	private StaticCostTable staticCostTable;
	private ReconstructionHostNode[] hostNodes;
	private ReconstructionParasiteNode[] parasiteNodes;
	private ArrayList<ReconstructionParasiteNode> parasiteFirsts;
	private ArrayList<ReconstructionParasiteNode> parasiteLasts;
	private ReconstructionCosts costs;
	private Map<String, String> options;
	private Map<InputHostNode, ReconstructionHostNode> hostRelations;
	private long calculationCounter;
	private Map<EventType, ReconstructionSolution> bestSolutions;
	private List<ReconstructionSolution> currentSolutions;

	public ReconstructionCophylogeny(InputCophylogeny cophylogeny) {
		this(cophylogeny, cophylogeny.getCosts(), cophylogeny.getOptions());
	}

	public ReconstructionCophylogeny(InputCophylogeny cophylogeny,
			ReconstructionCosts costs, Map<String, String> options) {
		setFilename(cophylogeny.getFilename());
		this.costs = costs;
		this.options = options;
		this.calculationCounter = 0;
		init(cophylogeny);
		if (this.options.get(OptionsConstants.PROBABILITY_COSTS_ID).equals(
				OptionsConstants.PROBABILITY_COSTS_1_ID))
			this.costs = CostCalculator.convertProbabilitiesToCosts(this.costs);
		this.staticCostTable = new StaticCostTable();
		this.staticCostTable.init(this.hostNodes,
				options.get(OptionsConstants.HOSTSWITCH_ID),
				options.get(OptionsConstants.SORTING_ID),
				options.get(OptionsConstants.TAKEOFF_ID),
				options.get(OptionsConstants.RANK_ID),
				costs.get(EventConstants.HOSTSWITCH),
				costs.get(EventConstants.SORTING));
		this.bestSolutions = new HashMap<EventType, ReconstructionSolution>();
		this.currentSolutions = new ArrayList<ReconstructionSolution>();

	}

	public void compute(OutputStreamWriter osw) {
		try {
			if (options.get(OptionsConstants.AUTOMATIC_COSTS_ID).equals(
					OptionsConstants.AUTOMATIC_COSTS_1_ID)) {
				int cycles = Integer.parseInt(options
						.get(OptionsConstants.RANDOM_CYCLES_ID));
				String seed = options.get(OptionsConstants.RANDOM_SEED_ID);
				Random rand = new Random();
				if (seed != null && !seed.equals("")) {
					try {
						rand = new Random(Long.parseLong(seed));
					} catch (Exception e) {
					}
				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_0_ID)) {
					for (int i = 0; i < cycles; i++) {
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
					}
				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_1_ID)) {
					// int simplexStartValueNumber = Math.min(10, cycles);
					double simplexXiValue = 0.1;
					double simplexFunctionTolerance = 0.00000000001;
					int simplexMaxIteration = 1000;
					for (int i = 0; i < cycles; i++) {
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
					}
					double[][] xi = {
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue } };
					TreeMap<Double, double[]> results = new TreeMap<Double, double[]>();
					for (ReconstructionSolution solution : this.bestSolutions
							.values()) {
						results.put(solution.getQuality().doubleValue(),
								solution.getDoubleCosts());
					}
					// simplexStartValueNumber = results.size();

					// new version
					Iterator<Entry<Double, double[]>> iter = results.entrySet()
							.iterator();
					while (iter.hasNext()) {
						Entry<Double, double[]> entry = iter.next();
						double[] initCosts = entry.getValue();
						NelderMead.search(initCosts, xi,
								simplexFunctionTolerance, this,
								simplexMaxIteration, new Info());
					}
					// // old version
					// for (int i = 0; i < simplexStartValueNumber; i++) {
					// double[] initCosts = results.firstEntry().getValue();
					// NelderMead.search(initCosts, xi,
					// simplexFunctionTolerance, this,
					// simplexMaxIteration, new Info());
					// results.remove(results.firstEntry().getKey());
					// }

				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_2_ID)) {
					// search
					// TODO
					System.out.println("SEARCH");
					for (int i = 0; i < cycles; i++) {
						// TODO
						System.out.print(i + " ");
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
					}
					// optimize
					// TODO
					System.out.println("\nOPTIMIZE");
					double simplexXiValue = 0.1;
					double simplexFunctionTolerance = 0.00000000001;
					int simplexMaxIteration = 1000;
					double[][] xi = {
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue } };
					Stack<ReconstructionSolution> solutions = new Stack<ReconstructionSolution>();
					for (ReconstructionSolution solution : getBestSolutions()
							.values()) {
						solutions.add(solution);
					}
					Set<EventType> handledEvents = new HashSet<EventType>();
					Set<EventType> knownEvents = getBestSolutions().keySet();
					for (int i = 0; i < solutions.size(); i++) {
						// TODO
						System.out.println(i + "/" + solutions.size());
						ReconstructionSolution currentSolution = solutions
								.get(i);
						EventType currentEvents = currentSolution.getEvents();
						while (!handledEvents.contains(currentEvents)) {
							RestrictedQualityCalculator rqc = new RestrictedQualityCalculator(
									currentEvents, knownEvents);
							double[] initCosts = currentSolution.getCosts()
									.getDoubleCosts();
							NelderMead.search(initCosts, xi,
									simplexFunctionTolerance, rqc,
									simplexMaxIteration, new Info());
							applyCosts(initCosts);
							solve(osw, false);
							Set<EventType> newCurrentEvents = new HashSet<EventType>();
							for (ReconstructionSolution solution : getCurrentSolutions()) {
								newCurrentEvents.add(solution.getEvents());
							}
							if (newCurrentEvents.contains(currentEvents)) {
								if (!handledEvents.contains(currentEvents))
									handledEvents.add(currentEvents);
								break;
							} else {
								knownEvents = getBestSolutions().keySet();
							}
						}

					}
				}
				ReconstructionSolution bestSolution = getBestSolution();
				if (bestSolution != null) {
					this.costs = bestSolution.getCosts();
					this.staticCostTable.update(
							this.costs.getHostswitchCosts(),
							this.costs.getSortingCosts());
					solve(osw, true);
				}
			} else {
				solve(osw, true);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void compute(OutputStreamWriter osw, IProgressMonitor monitor,
			boolean showResult) {
		try {
			if (options.get(OptionsConstants.AUTOMATIC_COSTS_ID).equals(
					OptionsConstants.AUTOMATIC_COSTS_1_ID)) {
				int cycles = Integer.parseInt(options
						.get(OptionsConstants.RANDOM_CYCLES_ID));
				String seed = options.get(OptionsConstants.RANDOM_SEED_ID);
				Random rand = new Random();
				if (seed != null && !seed.equals("")) {
					try {
						rand = new Random(Long.parseLong(seed));
					} catch (Exception e) {
					}
				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_0_ID)) {
					if (monitor != null)
						monitor.subTask("Compute " + cycles
								+ " reconstructions with random cost values");
					double workedFraction = 100d / (double) cycles;
					double worked = 0;
					int foundSolutions = 0;
					for (int i = 0; i < cycles
							&& (monitor == null || !monitor.isCanceled()); i++) {
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
						worked += workedFraction;
						if (monitor != null && worked > 1d) {
							int workedInteger = (int) Math.floor(worked);
							monitor.worked(workedInteger);
							worked -= workedInteger;
						}
						if (foundSolutions < getBestSolutions().size()) {
							foundSolutions = getBestSolutions().size();
							monitor.subTask("Found " + foundSolutions
									+ " distinct reconstructions");
						}
					}
					int[][] frequencies = new int[getBestSolutions().keySet()
							.size()][4];
					Iterator<EventType> iter = getBestSolutions().keySet()
							.iterator();
					for (int i = 0; i < frequencies.length; i++) {
						EventType events = iter.next();
						frequencies[i][0] = events.getCospeciations();
						frequencies[i][1] = events.getSortings();
						frequencies[i][2] = events.getDuplications();
						frequencies[i][3] = events.getHostswitches();

					}
				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_1_ID)) {
					// search
					if (monitor != null)
						monitor.subTask("Compute " + cycles
								+ " reconstructions with random cost values");
					double workedFraction = 50d / (double) cycles;
					double worked = 0;
					int foundSolutions = 0;
					for (int i = 0; i < cycles
							&& (monitor == null || !monitor.isCanceled()); i++) {
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
						worked += workedFraction;
						if (monitor != null && worked > 1d) {
							int workedInteger = (int) Math.floor(worked);
							monitor.worked(workedInteger);
							worked -= workedInteger;
						}
						if (foundSolutions < getBestSolutions().size()) {
							foundSolutions = getBestSolutions().size();
							monitor.subTask("Found " + foundSolutions
									+ " distinct reconstructions");
						}
					}
					// optimize
					int simplexStartValueNumber = Math.min(10, cycles);
					double simplexXiValue = 0.1;
					double simplexFunctionTolerance = 0.00000000001;
					int simplexMaxIteration = 1000;
					double[][] xi = {
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue } };
					TreeMap<Double, double[]> results = new TreeMap<Double, double[]>();
					for (ReconstructionSolution solution : this.bestSolutions
							.values()) {
						results.put(solution.getQuality().doubleValue(),
								solution.getDoubleCosts());
					}
					simplexStartValueNumber = results.size();
					workedFraction = 50d / (double) simplexStartValueNumber;
					worked = 0;
					Iterator<Entry<Double, double[]>> iter = results.entrySet()
							.iterator();
					int i = 0;
					while (iter.hasNext()
							&& (monitor == null || !monitor.isCanceled())) {
						if (monitor != null)
							monitor.subTask("Optimize "
									+ (i + 1)
									+ "/"
									+ simplexStartValueNumber
									+ " distinct reconstructions using Nelder-Mead's algorithm");
						Entry<Double, double[]> entry = iter.next();
						double[] initCosts = entry.getValue();
						NelderMead.search(initCosts, xi,
								simplexFunctionTolerance, this,
								simplexMaxIteration, new Info());
						worked += workedFraction;
						if (monitor != null && worked > 1d) {
							int workedInteger = (int) Math.floor(worked);
							monitor.worked(workedInteger);
							worked -= workedInteger;
						}
						i++;
					}
				}
				if (options.get(OptionsConstants.AUTOMATIC_METHOD_ID).equals(
						OptionsConstants.AUTOMATIC_METHOD_2_ID)) {
					// search
					if (monitor != null)
						monitor.subTask("Compute " + cycles
								+ " reconstructions with random cost values");
					double workedFraction = 50d / (double) cycles;
					double worked = 0;
					int foundSolutions = 0;
					for (int i = 0; i < cycles
							&& (monitor == null || !monitor.isCanceled()); i++) {
						double[] randomCosts = CostCalculator
								.computeRandomCosts(rand);
						applyCosts(randomCosts);
						solve(osw, false);
						worked += workedFraction;
						if (monitor != null && worked > 1d) {
							int workedInteger = (int) Math.floor(worked);
							monitor.worked(workedInteger);
							worked -= workedInteger;
						}
						if (foundSolutions < getBestSolutions().size()) {
							foundSolutions = getBestSolutions().size();
							monitor.subTask("Found " + foundSolutions
									+ " distinct reconstructions");
						}
					}
					// optimize
					double simplexXiValue = 0.1;
					double simplexFunctionTolerance = 0.00000000001;
					int simplexMaxIteration = 1000;
					double[][] xi = {
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue },
							{ simplexXiValue, simplexXiValue, simplexXiValue,
									simplexXiValue } };
					Stack<ReconstructionSolution> solutions = new Stack<ReconstructionSolution>();
					for (ReconstructionSolution solution : getBestSolutions()
							.values()) {
						solutions.add(solution);
					}
					workedFraction = 50d / (double) solutions.size();
					worked = 0;
					Set<EventType> handledEvents = new HashSet<EventType>();
					Set<EventType> knownEvents = getBestSolutions().keySet();
					for (int i = 0; i < solutions.size()
							&& (monitor == null || !monitor.isCanceled()); i++) {
						ReconstructionSolution currentSolution = solutions
								.get(i);
						EventType currentEvents = currentSolution.getEvents();
						while (!handledEvents.contains(currentEvents)) {
							RestrictedQualityCalculator rqc = new RestrictedQualityCalculator(
									currentEvents, knownEvents);
							double[] initCosts = currentSolution.getCosts()
									.getDoubleCosts();
							NelderMead.search(initCosts, xi,
									simplexFunctionTolerance, rqc,
									simplexMaxIteration, new Info());
							applyCosts(initCosts);
							solve(osw, false);
							Set<EventType> newCurrentEvents = new HashSet<EventType>();
							for (ReconstructionSolution solution : getCurrentSolutions()) {
								newCurrentEvents.add(solution.getEvents());
							}
							if (newCurrentEvents.contains(currentEvents)) {
								if (!handledEvents.contains(currentEvents))
									handledEvents.add(currentEvents);
								monitor.subTask("Optimized "
										+ handledEvents.size() + "/"
										+ solutions.size() + " solutions ("
										+ knownEvents.size() + " totaly found)");
								break;
							} else {
								knownEvents = getBestSolutions().keySet();
							}
						}

						worked += workedFraction;
						if (monitor != null && worked > 1d) {
							int workedInteger = (int) Math.floor(worked);
							monitor.worked(workedInteger);
							worked -= workedInteger;
						}
					}
				}

				ReconstructionSolution bestSolution = getBestSolution();
				if (bestSolution != null) {
					this.costs = bestSolution.getCosts();
					this.staticCostTable.update(
							this.costs.getHostswitchCosts(),
							this.costs.getSortingCosts());
					solve(osw, true);
				}
			} else {
				solve(osw, true);
				if (monitor != null) {
					monitor.worked(100);
				}
			}

			if (showResult) {
				PlatformUI.getWorkbench().getDisplay().syncExec(new Runnable() {

					public void run() {
						VisualOutputEditorInput input = new VisualOutputEditorInput(
								ReconstructionCophylogeny.this);
						try {
							PlatformUI.getWorkbench().getWorkbenchWindows()[0]
									.getActivePage().openEditor(input,
											VisualOutputEditor.ID, false);
							// PlatformUI.getWorkbench().getWorkbenchWindows()[0]
							// .getActivePage().showView(
							// NodePropertiesView.ID);
						} catch (PartInitException e) {
							CophylogenyLogger.logError(
									"Error while opening new editor", e);
						}

					}

				});
			}
		} catch (Exception e) {
			CophylogenyLogger.logError(
					"Error while computing the reconstruction", e);
		}
	}

	private void init(InputCophylogeny inputCophylogeny) {
		this.hostNodes = new ReconstructionHostNode[inputCophylogeny
				.getHostRoot().getChildrenCount() + 1];
		this.hostRelations = new HashMap<InputHostNode, ReconstructionHostNode>();
		initHostNodes(inputCophylogeny.getHostRoot(), null);
		this.setHostRoot(hostNodes[0]);
		this.parasiteNodes = new ReconstructionParasiteNode[inputCophylogeny
				.getParasiteRoot().getChildrenCount() + 1];
		this.parasiteFirsts = new ArrayList<ReconstructionParasiteNode>();
		this.parasiteLasts = new ArrayList<ReconstructionParasiteNode>();
		initParasiteNodes(inputCophylogeny.getParasiteRoot(), null);
		this.setParasiteRoot(parasiteNodes[0]);
	}

	private double solve(OutputStreamWriter osw, boolean printResults) {
		BigDecimal result = new BigDecimal(Double.MAX_VALUE);
		try {
			BaseSolver baseSolver = new BaseSolver();
			baseSolver.calculateMapping(this, this.staticCostTable);
			this.calculationCounter++;

			this.currentSolutions = new ArrayList<ReconstructionSolution>();
			List<Integer> bestRootMappings = computeBestRootMappings();

			for (Integer rootMapping : bestRootMappings) {
				if (rootMapping != null && rootMapping >= 0) {
					for (ReconstructionElement reconstruction : this.parasiteNodes[0]
							.getReconstructionInfo()[rootMapping]
							.getReconstructions()) {
						for (EventType events : reconstruction
								.getDerivedEvents()) {

							// TODO PARETOFRONT
							boolean insertEvents = true;
							Object[] currentEventsArray = this.currentSolutions
									.toArray();
							for (Object itemObj : currentEventsArray) {
								ReconstructionSolution itemSolution = (ReconstructionSolution) itemObj;
								EventType item = itemSolution.getEvents();
								if (item.getCospeciations() <= events
										.getCospeciations()
										&& item.getSortings() <= events
												.getSortings()
										&& item.getDuplications() <= events
												.getDuplications()
										&& item.getHostswitches() <= events
												.getHostswitches()) {
									insertEvents = false;
									break;
								}
								if (item.getCospeciations() >= events
										.getCospeciations()
										&& item.getSortings() >= events
												.getSortings()
										&& item.getDuplications() >= events
												.getDuplications()
										&& item.getHostswitches() >= events
												.getHostswitches()) {
									this.currentSolutions.remove(itemSolution);
								}
							}
							if (!insertEvents)
								continue;

							BigDecimal quality = QualityCalculator
									.computeQuality(this.costs, events);
							ReconstructionSolution solution = new ReconstructionSolution(
									rootMapping, reconstruction.getCosts(),
									this.costs, quality, events);
							this.currentSolutions.add(solution);

							if (quality.compareTo(BigDecimal.ZERO) >= 0) {
								if (quality.compareTo(result) < 0) {
									result = quality;
								}
								if (!this.bestSolutions.containsKey(events)
										|| this.bestSolutions.get(events)
												.getQuality()
												.compareTo(quality) > 0) {
									this.bestSolutions.put(events, solution);
								}
							}
						}
					}
				}
			}
			if (printResults) {
				printReconstructionInfo(getBestCurrentSolution(), osw);
			}
			// System.gc();
		} catch (Exception e) {
			CophylogenyLogger.logError(e);
		}
		return result.doubleValue();
	}

	public void applyCosts(double[] costs) {
		this.costs.setCospeciationCosts(new BigDecimal(costs[0]));
		this.costs.setSortingCosts(new BigDecimal(costs[1]));
		this.costs.setDuplicationCosts(new BigDecimal(costs[2]));
		this.costs.setHostswitchCosts(new BigDecimal(costs[3]));
		this.staticCostTable.update(this.costs.getHostswitchCosts(),
				this.costs.getSortingCosts());
	}

	public void applyCosts(ReconstructionCosts costs) {
		this.costs = costs;
		this.staticCostTable.update(this.costs.getHostswitchCosts(),
				this.costs.getSortingCosts());
	}

	/**
	 * compute the best mapping for the parasite root node to a host node
	 * 
	 * @return a list of the best host mapping partners
	 */
	public List<Integer> computeBestRootMappings() {
		List<Integer> bestRootMappings = new ArrayList<Integer>();
		bestRootMappings.add(0);
		if (!this.options.get(OptionsConstants.ROOT_MAPPING_ID).equals(
				OptionsConstants.ROOT_MAPPING_1_ID)) {
			BigDecimal costs = null;
			for (ReconstructionInfo reconstructionInfo : this.parasiteNodes[0]
					.getReconstructionInfo()) {
				if (reconstructionInfo.getCosts() != null) {
					if (costs == null
							|| reconstructionInfo.getCosts().compareTo(costs) < 0) {
						costs = reconstructionInfo.getCosts();
						bestRootMappings = new ArrayList<Integer>();
						bestRootMappings.add(reconstructionInfo.getHostIndex());
					} else if (reconstructionInfo.getCosts().compareTo(costs) == 0) {
						bestRootMappings.add(reconstructionInfo.getHostIndex());
					}
				}
			}
		}
		return bestRootMappings;
	}

	/**
	 * initializes the host nodes for the reconstruction according to the given
	 * input host nodes
	 * 
	 * @param inputHostNode
	 * @param parent
	 * @return
	 */
	private ReconstructionHostNode initHostNodes(InputHostNode inputHostNode,
			ReconstructionHostNode parent) {
		ReconstructionHostNode node = new ReconstructionHostNode(inputHostNode);
		node.setParent(parent);

		for (InputHostNode childNode : inputHostNode.getChildren()) {
			node.getChildren().add(initHostNodes(childNode, node));
		}

		this.hostNodes[node.getNumber()] = node;
		this.hostRelations.put(inputHostNode, node);
		return node;
	}

	/**
	 * initializes the parasite nodes for the reconstruction according to the
	 * given input parasite nodes data
	 * 
	 * @param inputParasiteNode
	 * @param parent
	 * @return
	 */
	private ReconstructionParasiteNode initParasiteNodes(
			InputParasiteNode inputParasiteNode,
			ReconstructionParasiteNode parent) {
		ReconstructionParasiteNode node = new ReconstructionParasiteNode(
				inputParasiteNode);
		node.setParent(parent);
		if (this.parasiteLasts.size() <= node.getLevel()) {
			this.parasiteFirsts.add(node);
			this.parasiteLasts.add(node);
		} else {
			node.setPrev(this.parasiteLasts.get(node.getLevel()));
			this.parasiteLasts.get(node.getLevel()).setNext(node);
			this.parasiteLasts.set(node.getLevel(), node);
		}

		for (InputParasiteNode childNode : inputParasiteNode.getChildren()) {
			node.getChildren().add(initParasiteNodes(childNode, node));
		}

		for (InputHostNode hostNode : inputParasiteNode.getAssociations()) {
			node.getAssociations().add(this.hostRelations.get(hostNode));
		}

		this.parasiteNodes[node.getNumber()] = node;
		return node;
	}

	/**
	 * @return the parasiteFirsts
	 */
	public ArrayList<ReconstructionParasiteNode> getParasiteFirsts() {
		return parasiteFirsts;
	}

	/**
	 * @return the hostNodes
	 */
	public ReconstructionHostNode[] getHostNodes() {
		return hostNodes;
	}

	/**
	 * @return the parasiteNodes
	 */
	public ReconstructionParasiteNode[] getParasiteNodes() {
		return parasiteNodes;
	}

	/**
	 * @return the costs
	 */
	public ReconstructionCosts getCosts() {
		return costs;
	}

	/**
	 * @return the options
	 */
	public Map<String, String> getOptions() {
		return options;
	}

	/**
	 * writes the reconstruction info to an OutputStreamWriter or System.out
	 * 
	 * @param quality
	 * @param events
	 * @param rootMapping
	 * @param ow
	 */
	private void printReconstructionInfo(ReconstructionSolution solution,
			OutputStreamWriter ow) {
		ReconstructionCosts costs = solution.getCosts();
		ReconstructionCosts probabilities = CostCalculator
				.convertCostsToProbabilities(costs);
		EventType events = solution.getEvents();
		int eventSum = events.getEventSum();
		if (ow != null) {
			try {
				ow.write(String.valueOf(this.calculationCounter));
				ow.write("\t");
				ow.write(solution.getQuality().toPlainString());
				ow.write("\t");
				ow.write(solution.getTotalCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(eventSum));
				ow.write("\t");
				ow.write(costs.getCospeciationCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getCospeciations()));
				ow.write("\t");
				ow.write(probabilities.getCospeciationCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getCospeciations()
						/ (double) eventSum));
				ow.write("\t");
				ow.write(costs.getSortingCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getSortings()));
				ow.write("\t");
				ow.write(probabilities.getSortingCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getSortings()
						/ (double) eventSum));
				ow.write("\t");
				ow.write(costs.getDuplicationCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getDuplications()));
				ow.write("\t");
				ow.write(probabilities.getDuplicationCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getDuplications()
						/ (double) eventSum));
				ow.write("\t");
				ow.write(costs.getHostswitchCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getHostswitches()));
				ow.write("\t");
				ow.write(probabilities.getHostswitchCosts().toPlainString());
				ow.write("\t");
				ow.write(String.valueOf(events.getHostswitches()
						/ (double) eventSum));
				ow.write("\t");
				ow.write(String.valueOf(solution.getRootMapping()));
				ow.write("\n");
			} catch (IOException e) {
				CophylogenyLogger.logError(e);
			}
		} else {
			System.out.print(this.calculationCounter);
			System.out.print("\t");
			System.out.print(solution.getQuality().toPlainString());
			System.out.print("\t");
			System.out.print(solution.getTotalCosts().toPlainString());
			System.out.print("\t");
			System.out.print(eventSum);
			System.out.print("\t");
			System.out.print(costs.getCospeciationCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getCospeciations());
			System.out.print("\t");
			System.out.print(probabilities.getCospeciationCosts()
					.toPlainString());
			System.out.print("\t");
			System.out.print(events.getCospeciations() / (double) eventSum);
			System.out.print("\t");
			System.out.print(costs.getSortingCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getSortings());
			System.out.print("\t");
			System.out.print(probabilities.getSortingCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getSortings() / (double) eventSum);
			System.out.print("\t");
			System.out.print(costs.getDuplicationCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getDuplications());
			System.out.print("\t");
			System.out.print(probabilities.getDuplicationCosts()
					.toPlainString());
			System.out.print("\t");
			System.out.print(events.getDuplications() / (double) eventSum);
			System.out.print("\t");
			System.out.print(costs.getHostswitchCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getHostswitches());
			System.out.print("\t");
			System.out
					.print(probabilities.getHostswitchCosts().toPlainString());
			System.out.print("\t");
			System.out.print(events.getHostswitches() / (double) eventSum);
			System.out.print("\t");
			System.out.print(solution.getRootMapping());
			System.out.println();
		}
	}

	/**
	 * @return get a map of the distinct reconstructions (different number of
	 *         occurring events) with best quality value ever computed (i.e.
	 *         computed with different cost vector)
	 */
	public Map<EventType, ReconstructionSolution> getBestSolutions() {
		return this.bestSolutions;
	}

	/**
	 * @return get the quality value of the best reconstruction
	 */
	private ReconstructionSolution getBestSolution() {
		ReconstructionSolution bestSolution = null;
		for (ReconstructionSolution solution : this.bestSolutions.values()) {
			if (bestSolution == null
					|| bestSolution.getQuality().compareTo(BigDecimal.ZERO) < 0
					|| (solution.getQuality().compareTo(BigDecimal.ZERO) >= 0 && bestSolution
							.getQuality().compareTo(solution.getQuality()) > 0)) {
				bestSolution = solution;
			}
		}
		return bestSolution;
	}

	/**
	 * @return get a list of the reconstructions with minimal costs, computed
	 *         during the last computation step (i.e. computed with the last
	 *         cost vector)
	 */
	public List<ReconstructionSolution> getCurrentSolutions() {
		return this.currentSolutions;
	}

	/**
	 * @return get the quality value of the current reconstruction
	 */
	private ReconstructionSolution getBestCurrentSolution() {
		ReconstructionSolution bestCurrentSolution = null;
		for (ReconstructionSolution solution : this.currentSolutions) {
			if (bestCurrentSolution == null
					|| bestCurrentSolution.getQuality().compareTo(
							BigDecimal.ZERO) < 0
					|| (solution.getQuality().compareTo(BigDecimal.ZERO) >= 0 && bestCurrentSolution
							.getQuality().compareTo(solution.getQuality()) > 0)) {
				bestCurrentSolution = solution;
			}
		}
		return bestCurrentSolution;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * de.jtem.numericalMethods.calculus.function.RealFunctionOfSeveralVariables
	 * #eval(double[])
	 */

	public double eval(double[] costs) {
		if (!CostCalculator.isValidCostVector(costs)) {
			return Double.MAX_VALUE;
		}
		// TODO change to generate cost sets with sum = 1
		double sum = costs[0] + costs[1] + costs[2] + costs[3];
		costs[0] = costs[0] / sum;
		costs[1] = costs[1] / sum;
		costs[2] = costs[2] / sum;
		costs[3] = costs[3] / sum;
		applyCosts(costs);
		return solve(null, false);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * de.jtem.numericalMethods.calculus.function.RealFunctionOfSeveralVariables
	 * #getNumberOfVariables()
	 */

	public int getNumberOfVariables() {
		return 4;
	}

}
